
#define GORING0_DLL

#include <windows.h>
#include <winbase.h>

#include "constants.h"
#include "goring0.h"

unsigned int user_stack;

/**
 * Switch from ring 3 to ring 0
 */
__declspec(dllexport) __declspec(naked) void GoRing0(){
	
	_asm {
		//Save the userland esp
		mov user_stack, esp
		
		//Call the interrupt handler
		int INT_VECT_TO_HOOK
		
		//A stack switch has occured. The stack is not the same anymore. 
		//Move back to the old stack: the userland one, but in kernel mode!
		mov esp, user_stack
		
		ret
	}
}

/**
 * Switch from ring 0 to ring 3
 */
__declspec(dllexport) __declspec(naked) void GoRing3(){
	
	_asm {
	
		//Since we'll go back to ring 3, there will be a stack switch.
		//But we now are in ring 0, so esp and ss won't be pushed.
		//We need to push them so that the stack switch doesn't crash everything
		mov eax, esp
		push 23h				//push userland ss
		push eax				//push the value that esp would have had before the int with stack switch
		
		//Call the interrupt handler
		int INT_VECT_TO_HOOK
		
		//Restore FS to its userland value (has been smashed by iret)
		push 0x3b
		pop fs
		
		ret
	}
}


BOOL WINAPI DllMain(HANDLE hinstDLL, DWORD dwReason, LPVOID lpvReserved){
	return TRUE;
}

